# python 3.6
# -*- coding: utf-8 -*-
"""
Created on June 11, 17:16:19, 2018
@author: Wade Roberts

For full documentation, see https://waderoberts123.github.io/Hydrostats/
"""
from hydrostats.metrics import *
from hydrostats.analyze import *
